// Copyright 2013-2025 Daniel Parker
// Distributed under Boost license

#include <jsoncons/json.hpp>
#include <catch/catch.hpp>

using namespace jsoncons;

TEST_CASE("detail::from_integer tests")
{
    SECTION("")
    {
    }
}

