// Copyright 2013-2025 Daniel Parker
// Distributed under the Boost license, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

// See https://github.com/danielaparker/jsoncons for latest version

#ifndef JSONCONS_EXT_JSONSCHEMA_JSONSCHEMA_HPP
#define JSONCONS_EXT_JSONSCHEMA_JSONSCHEMA_HPP

#include <jsoncons_ext/jsonschema/json_schema_factory.hpp>

#endif // JSONCONS_EXT_JSONSCHEMA_JSONSCHEMA_HPP
